// -*-c++-*-

/***************************************************************************
                          rccvector2d.h  -  Interface for an all purpose
			                    two dimensional geometric vector
					    class.
                             -------------------
    begin                : 17-JULY-2001
    copyright            : (C) 2001 by Tom Howard.
    email                : thomashoward@email.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef _RCC_VECTOR2D_H_
#define _RCC_VECTOR2D_H_

#pragma interface

#include "rccobject.h"
#include "rccmemory.h"

#include "rccangle.h"
#include <math.h>
#include "rccexception.h"

class RCC_Vector2d : public RCC_Object
{
public:
  enum mode_t { POLAR, CART, BOTH };

  RCC_Vector2d ();
  RCC_Vector2d ( const double&, const RCC_Angle& );
  RCC_Vector2d ( const double&, const double& );
  
  ~RCC_Vector2d();

  RCC_Pointer < RCC_Object > copy () const { return new RCC_Vector2d ( *this ); }

  double dist() const 
  { 
    if ( valid ) 
      {
	try 
	  {
	  toPolar (); 
	  return distance;
	  } 
	catch ( ... ) 
	  {
	  }
      }
    return HUGE_VAL; 
  }

  RCC_Angle dir() const 
  { 
    if ( valid ) 
      { 
	try {
	  toPolar (); 
	  return direction; 
	} catch ( ... ) {}
      } 
    return RCC_Angle (); 
  }

  double mag() const { return dist(); }
  RCC_Angle head() const { return dir(); }

  double x() const 
  { 
    if ( valid ) 
      { 
	try 
	  { 
	  toCart (); 
	  return xval; 
	  } 
	catch ( ... ) 
	  {
	  }
      } 
    return HUGE_VAL; 
  }

  double y() const 
  { 
    if ( valid ) 
      { 
	try {
	  toCart (); 
	  return yval; 
	} catch ( ... ) {} 
      } 
    return HUGE_VAL; 
  }

  RCC_Vector2d& rotate(const RCC_Angle& ang)
  {
    if ( valid )
      {
	if ( ang.isValid() )
	  {
	    try 
	      {
		toPolar();
		if ( direction.isValid () )
		  {
		    direction += ang;
		    mode = POLAR;
		  }
	      }
	    catch ( ... ) 
	      {
		valid = false;
	      }
	  }
	else
	  direction.clear () ;
      }
    return *this;
  }

  const RCC_Vector2d& norm() const
  {
    if ( valid && mode != RCC_Vector2d::CART )
      {
	if ( distance < 0.0 )
	  {
	    distance = -distance;
	    direction += RCC_Angle(M_PI, RCC_Angle::RAD);
	  }
      }
    return *this;
  }

  bool isValid() const { return valid; }
  
  RCC_Vector2d& clear () { valid = false; return *this; }

  RCC_Vector2d operator+() const { return *this; }
  RCC_Vector2d operator-() const { return (*this * -1.0); }

  RCC_Vector2d& operator+=(const RCC_Vector2d& i);
  RCC_Vector2d& operator-=(const RCC_Vector2d& i);

  RCC_Vector2d& operator*=(const double& scaler);
  RCC_Vector2d& operator/=(const double& scaler);
    
  friend RCC_Vector2d& operator+(RCC_Vector2d a, const RCC_Vector2d& b) { return a += b; }
  friend RCC_Vector2d& operator-(RCC_Vector2d a, const RCC_Vector2d& b) { return a -= b; }
  friend RCC_Vector2d& operator*(RCC_Vector2d v, const double& scaler) { return v *= scaler; }
  friend RCC_Vector2d& operator*(const double& scaler, RCC_Vector2d v) { return v *= scaler; }
  friend RCC_Vector2d& operator/(RCC_Vector2d v, const double& scaler) { return v /= scaler; }
  friend bool operator==(const RCC_Vector2d& a, const RCC_Vector2d& b) { return equals (a, b); }
  friend bool operator!=(const RCC_Vector2d& a, const RCC_Vector2d& b) { return !(a == b); }

    
  friend std::ostream& operator<< ( std::ostream& o, const RCC_Vector2d& v) { return v.print ( o ); }

  std::ostream& print ( std::ostream& o ) const;
  std::ostream& print ( std::ostream& o, RCC_Vector2d::mode_t vm, RCC_Angle::mode_t ) const;
  
  friend bool equals(const RCC_Vector2d& a, const RCC_Vector2d& b); 
 
private:
  bool valid;
  mutable double distance;
  mutable RCC_Angle direction;

  mutable double xval, yval;

  mutable mode_t mode;

  const RCC_Vector2d& toPolar() const
  {
    if ( valid && mode == RCC_Vector2d::CART )
      {
	if ( xval == HUGE_VAL )
	  throw RCC_GeomException ( "Can not convert RCC_Vector2d to polar if x value is unknown" );
	if ( yval == HUGE_VAL )
	  throw RCC_GeomException ( "Can not convert RCC_Vector2d to polar if y value is unknown" );

	distance = hypot ( xval, yval );
	direction = RCC_Angle::atan2 ( yval, xval );
	mode = RCC_Vector2d::BOTH;
      }
    return *this;
  }

  const RCC_Vector2d& toCart() const
  {
    if ( valid && mode == RCC_Vector2d::POLAR )
      {
	if ( distance == HUGE_VAL )
	  throw RCC_GeomException ( "Can not convert RCC_Vector2d to cartesian if distance is unknown" );
	if ( !direction.isValid () )
	  throw RCC_GeomException ( "Can not convert RCC_Vector2d to cartesian if direction is unknown" );
	
	yval = distance * direction.sin ();
	xval = distance * direction.cos ();
	mode = BOTH;
      }
    return *this;
  }


};

#endif
